/*
 * Copyright (c) 2007, Gregory Fleischer (gfleischer@gmail.com)
 * 
 * All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *   3. The names of the authors may not be used to endorse or promote
 *      products derived from this software without specific prior
 *      written permission.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/*
 * $Id: tftpgrab.c 239 2007-09-18 00:19:20Z gregory $
 */

/*
 TFTP specification - RFCs: 1350, 2347, 2348, 2349 
*/

#include "tftpgrab.h"

int verbose = 0;
int debuglevel = 0;
int dump_packet = 0;
int console_print = 0;
int file_overwrite = 0;
int include_filename = 1;
int check_bad_checksums = 0;

void
usage(const char *progname)
{
  fprintf(stdout, "Usage: %s [OPTION]... [-r FILE] [EXPRESSION]\n", progname);
  fprintf(stdout, "Reconstruct TFTP file contents from PCAP capture file.\n");
  fprintf(stdout, "With no FILE, or when FILE is -, read standard input.\n");
  fprintf(stdout, "   -r       PCAP file to read\n");
  fprintf(stdout, "   -f       overwrite existing files\n");
  fprintf(stdout, "   -c       print TFTP file contents to console\n");
  fprintf(stdout, "   -E       exclude TFTP filename when reconstructing\n");
  fprintf(stdout, "   -v       print verbose TFTP exchanges (repeat up to three times)\n");
  fprintf(stdout, "   -X       dump TFTP packet contents\n");
  fprintf(stdout, "   -B       check packets for bad checksums\n");
  fprintf(stdout, "   -d       specify debugging level\n");

}

void
debug(int level, const char *fmt, ...)
{
  va_list ap;

  if (level > debuglevel) 
    return;

  va_start(ap, fmt);
  fprintf(stderr, "[debug] ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);

}

void
warning(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "[warning] ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
}

void
die(int rc, const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "[error] ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);

  if (0 != rc) {
    exit(rc);
  }

}

void
hexdump(const u_char *buf, u_int len)
{
  char hexbuf[40 + 1], *hptr;
  char ascbuf[16 + 1], *aptr;
  u_short s1, s2;
  u_int i = 0;
  u_int output = 0;
  u_int count = len / sizeof(u_short);
  const u_char *bptr = buf;

  memset(hexbuf, 0, sizeof(hexbuf));
  memset(ascbuf, 0, sizeof(ascbuf));

  hptr = hexbuf;
  aptr = ascbuf;
  while (count > 0) {
    s1 = *(bptr++);
    s2 = *(bptr++);
    snprintf(hptr, 6, " %02x%02x", s1, s2);
    hptr += 5;
    *(aptr++) = isgraph(s1) ? s1 : '.';
    *(aptr++) = isgraph(s2) ? s2 : '.';
    ++i;
    if (i >= 8) {
      printf("\t0x%04X: %s  %s\n", output, hexbuf, ascbuf);
      output += 16;
      hptr = hexbuf;
      aptr = ascbuf;
      i = 0;
      if ((count > 8) && (0 == memcmp(bptr - 16, bptr, 16))) {
	printf("\t*\n");
	do {
	  bptr += 16;
	  output += 16;
	  count -= 8;
	} while ((count > 8) && (0 == memcmp(bptr - 16, bptr, 16)));
      } 
    }
    --count;
  }
  if (len % 2) {
    s1 = *bptr++;
    snprintf(hptr, 6, " %02x  ", s1);
    hptr += 5;
    *(aptr++) = isgraph(s1) ? s1 : '.';
    *(aptr++) = ' ';
    ++i;
  }
  if (0 != i) {
    while (i < 8) {
      strcat(hptr, "     ");
      hptr += 5;
      *(aptr++) = ' ';
      *(aptr++) = ' ';
      ++i;
    }
    printf("\t0x%04X: %s  %s\n", output, hexbuf, ascbuf);
  } else {
    printf("\t0x%04X:\n", output);
  }
  fflush(stdout);
}

u_short
checksum(const u_char *buffer, u_int len, u_int32_t cksum)
{
  /* see RFC 1071 for IP checksum algorithm*/
  const u_short *addr = (const u_short *) buffer;
  u_int32_t sum = cksum;

  while (len > 1) {
    sum += *addr++;
    len -= 2;
  }
  if (len > 0) {
    sum += htons((*(u_char*)addr) << 8);
  }
  while (sum >> 16) {
    sum = (sum & 0xffff) + (sum >> 16);
  }
  return (u_short) ~sum;
}

#if !defined(IPOPT_EOL)
#define IPOPT_EOL  0x0
#define IPOPT_NOP  0x1
#define IPOPT_LSRR 0x83 /* loose routing */
#define IPOPT_SSRR 0x89 /* strict routing */
#endif

u_int32_t
find_source_route(const struct ip *iph)
{
  const u_char *optptr = (const u_char *)(iph + 1);
  u_int optlen = (iph->ip_hl << 2) - sizeof(struct ip);

  const u_char *c;
  u_int pos = 0;
  u_int len = 0;
  u_int ptr = 0;
  u_int32_t route_dst = 0;
  
  c = optptr;
  while (pos < optlen) {
    
    if (IPOPT_NOP == *c) {
      ++c; ++pos;
      continue;
    } else if (IPOPT_EOL == *c) {
      break;
    }

    len = (u_int)c[1];
    if ( (len < 2) || (len > optlen) || ((pos + len) > optlen) ){
      break;
    }

    switch (*c) {
    case IPOPT_LSRR:
    case IPOPT_SSRR:
      if (len < 7) {
	goto done;
      }
      ptr = (u_int)c[2];
      if (ptr < 4 || (0 != (ptr % 4))) {
	goto done;
      }
      if (!(ptr > len)) {
	memcpy(&route_dst, c + len - sizeof(u_int32_t), sizeof(u_int32_t));
      }
      goto done;
    }
    c += len;
    pos += len;
      
  }

 done:
  if (0 == route_dst) {
    return iph->ip_dst.s_addr;
  } else {
    return route_dst;
  }
  
}

u_short
udp_checksum(const struct ip *iph, const struct udphdr *udph, u_int ulen)
{
  /* see RFC 768 for UDP checksum algorithm */
  u_int32_t cksum = 0;
  u_int32_t src = iph->ip_src.s_addr;
  u_int32_t dst = iph->ip_dst.s_addr;

  /* check for source routing */
  if (iph->ip_hl > 5) {
    dst = find_source_route(iph);
  }

  cksum += (src & 0xffff0000) >> 16;
  cksum += src & 0xffff;
  cksum += (dst & 0xffff0000) >> 16;
  cksum += dst & 0xffff;
  cksum += htons(IPPROTO_UDP);
  cksum += udph->uh_ulen;

  return checksum((const u_char *)udph, ulen, cksum);
}

typedef void (*PACKET_HANDLER)(u_char *, const struct pcap_pkthdr *, const u_char *);


typedef struct ip_fragment_data {
  struct ip_fragment_data *next;
  u_int offset;
  size_t datalen;
  u_char *data;
} ip_fragment_data_t;

typedef struct ip_fragment {
  struct ip_fragment *next;
  struct ip header;
  u_short flags;
  size_t datalen;
  ip_fragment_data_t *data;
  u_char *packetdata;
} ip_fragment_t;

#define FIRST_FRAGMENT 0x1
#define LAST_FRAGMENT  0x2
#define REASSEMBLED_FRAGMENT  0x4
#define BAD_FRAGMENT  0x8

#define HASH_SIZE 101
ip_fragment_t *ip_fragments[HASH_SIZE];

ushort
ip_fragment_hashcode(u_int32_t src, u_int32_t dst, u_short ip_id)
{
  ushort hashcode = 0;
  do {
    hashcode = ((hashcode << 8) + (src & 0xff)) % HASH_SIZE;
  } while (src >>= 8);
  do {
    hashcode = ((hashcode << 8) + (dst & 0xff)) % HASH_SIZE;
  } while (dst >>= 8);
  do {
    hashcode = ((hashcode << 8) + (ip_id & 0xff)) % HASH_SIZE;
  } while (ip_id >>= 8);

  return hashcode;
}

ip_fragment_t*
find_ip_fragment(const struct ip *header)
{
  u_int32_t src = header->ip_src.s_addr;
  u_int32_t dst = header->ip_dst.s_addr;
  u_short ip_id = header->ip_id;
  ip_fragment_t *ptr = NULL;
  u_short hashcode = 0;

  hashcode = ip_fragment_hashcode(src, dst, ip_id);
  ptr = ip_fragments[hashcode];

  while (ptr) {
    if ((ip_id == ptr->header.ip_id) &&
	(src == ptr->header.ip_src.s_addr) &&
	(dst == ptr->header.ip_dst.s_addr)) {
      return ptr;
    }
    ptr = ptr->next;
  }
  return NULL;
}

ip_fragment_t*
insert_ip_fragment(const struct ip *header)
{
  u_int32_t src = header->ip_src.s_addr;
  u_int32_t dst = header->ip_dst.s_addr;
  u_short ip_id = header->ip_id;
  ip_fragment_t *ptr = NULL;
  u_short hashcode = 0;

  ip_fragment_t *newfrag = (ip_fragment_t*)(malloc(sizeof(ip_fragment_t)));
  if (NULL == newfrag) {
    die(-1, "failed to allocate new fragment");
  }

  memset(newfrag, 0, sizeof(ip_fragment_t));

  hashcode = ip_fragment_hashcode(src, dst, ip_id);
  ptr = ip_fragments[hashcode];

  newfrag->next = ptr;
  ip_fragments[hashcode] = newfrag;

  return newfrag;
}					    

ip_fragment_data_t*
alloc_fragment_data(u_int offset, const u_char *data, size_t datalen)
{

  ip_fragment_data_t *newdata;

  if ( NULL == (newdata = (ip_fragment_data_t*)(malloc(sizeof(ip_fragment_data_t)))) ) {
    die(-1, "failed to allocate new fragment data");
  }
  if ( NULL == (newdata->data = (u_char*)(malloc(datalen))) ) {
    die(-1, "failed to allocate new fragment data of length: %d", datalen);
  }
  newdata->next = 0;
  newdata->offset = offset;
  newdata->datalen = datalen;
  memcpy(newdata->data, data, datalen);

  return newdata;
}

void
delete_ip_fragment_data(ip_fragment_t *frag)
{
  ip_fragment_data_t *tmp, *ptr = frag->data;
  while (ptr) {
    tmp = ptr;
    ptr = ptr->next;
    free(tmp->data);
    free(tmp);
  }
  frag->data = NULL;
}

void
add_ip_fragment_data(ip_fragment_t *frag, u_int offset, const u_char *data, size_t datalen)
{
  /* insert into list of fragments ordered by offset */
  ip_fragment_data_t *prev = 0, *ptr = frag->data;

  if (NULL == ptr) {
    frag->data = alloc_fragment_data(offset, data, datalen);
  } else {
    while (ptr) {
      if (offset < ptr->offset) {
	/* insert into list */
	ip_fragment_data_t *newdata = alloc_fragment_data(offset, data, datalen);
	newdata->next = ptr;
	if (0 == prev) {
	  frag->data = newdata;
	} else {
	  prev->next = newdata;
	}
	return;
      } 
      if ((offset < (ptr->offset + ptr->datalen)) &&
	  ((offset + datalen) <= (ptr->offset + ptr->datalen))) {
	/* contained within another fragment*/
	return; 
      }
      prev = ptr;
      ptr = ptr->next;
    }
    /* add to end of list */
    prev->next = alloc_fragment_data(offset, data, datalen);
  }

}


ip_fragment_t *
add_ip_fragment(const struct ip *header, u_int offset, u_short morefrags, const u_char *data, u_int caplen)
{
  size_t len;
  void *ptr;
  ip_fragment_t *frag = find_ip_fragment(header);

  if (NULL == frag) {
    /* first seen */
    frag = insert_ip_fragment(header);
    frag->header = *header;
  } else if (0 == offset) {
    frag->header = *header;
  }

  if (frag->flags & BAD_FRAGMENT) {
    return NULL;
  }

  offset <<= 3;
  len = offset + caplen;

  if (len > IP_MAXPACKET) {
    warning("length of fragment packet exceeds max possible data; offset: %d, caplen: %d, length: %d", offset, caplen, IP_MAXPACKET);
    delete_ip_fragment_data(frag);
    frag->flags |= BAD_FRAGMENT;
    return NULL;
  }

  if (0 == offset) {
    frag->flags |= FIRST_FRAGMENT;
  }
  if (!morefrags) {
    frag->flags |= LAST_FRAGMENT;
    frag->datalen = len;
  }

  add_ip_fragment_data(frag, offset, data, caplen);

  return frag;
}

int
ip_fragment_complete(ip_fragment_t *frag)
{

  u_char *packetdata;
  ip_fragment_data_t *ptr, *tmp;
  u_int offset, next_offset;

  if (frag->flags & REASSEMBLED_FRAGMENT) {
    warning("fragment [%p] was already reassembled", frag);
    return 0;
  }
  if (frag->flags & BAD_FRAGMENT) {
    warning("fragment [%p] was bad", frag);
    return 0;
  }

  if ((frag->flags & FIRST_FRAGMENT) && (frag->flags & LAST_FRAGMENT)) {

    ptr = frag->data;
    offset = next_offset = 0;
    while (ptr) {
      if (ptr->offset <= offset) {
	next_offset = ptr->offset + ptr->datalen;
	if (next_offset > offset) {
	  offset = next_offset;
	}
      } else {
	break;
      }
      ptr = ptr->next;
    }

    if (offset < frag->datalen) {
      return 0;
    } else if (offset > frag->datalen) {
      delete_ip_fragment_data(frag);
      frag->flags |= BAD_FRAGMENT;
      return 0;
    }

    if ( NULL == (packetdata = malloc(frag->datalen))) {
      die(2, "failed to allocate packet data length: %d", frag->datalen);
    }

    ptr = frag->data;
    offset = next_offset = 0;
    while (ptr) {
      if (ptr->offset <= offset) {
	next_offset = ptr->offset + ptr->datalen;
	if (next_offset > offset) {
	  memcpy(packetdata + offset, ptr->data, ptr->datalen);
	  offset = next_offset;
	}
      }
      tmp = ptr;
      ptr = ptr->next;
      /* release memory */
      free(tmp->data);
      free(tmp);
    }

    frag->packetdata = packetdata;
    frag->flags |= REASSEMBLED_FRAGMENT;
    return 1;
  }
  return 0;
}

void
remove_ip_fragment(ip_fragment_t *frag)
{
  u_int32_t src = frag->header.ip_src.s_addr;
  u_int32_t dst = frag->header.ip_dst.s_addr;
  u_short ip_id = frag->header.ip_id;
  ip_fragment_t *ptr = NULL;
  u_short hashcode = 0;

  hashcode = ip_fragment_hashcode(src, dst, ip_id);
  ptr = ip_fragments[hashcode];

  if (ptr == frag) {
    ip_fragments[hashcode] = frag->next;
  } else {
    while (ptr) {
      if (frag == ptr->next) {
	ptr->next = frag->next;
	break;
      }
      ptr = ptr->next;
    }
  }

  if (NULL == ptr) {
    debug(2, "failed to remove frag: %p", frag);
  } else {
    if (frag->flags & REASSEMBLED_FRAGMENT) { 
      free(frag->packetdata);
    }
    free(frag);
  }

}

void
cleanup_ip_fragments()
{
  int i;
  ip_fragment_t *tmp, *ptr = NULL;
  ip_fragment_data_t *data;
  char srcaddr[INET_ADDRSTRLEN], dstaddr[INET_ADDRSTRLEN];

  for (i = 0; i < HASH_SIZE; ++i) {
    ptr = ip_fragments[i];
    while (ptr) {
      data = ptr->data;
      if (data && (debuglevel >= 2)) {
	/* dump debuf info about outstanding fragments */
	if (inet_ntop(AF_INET, &(ptr->header.ip_src), srcaddr, sizeof(srcaddr)) == NULL) {
	  strcpy(srcaddr, "?");
	}
	if (inet_ntop(AF_INET, &(ptr->header.ip_dst), dstaddr, sizeof(dstaddr)) == NULL) {
	  strcpy(dstaddr, "?");
	}
	debug(2, "fragment bucket %d: [%p], length=%u, flags=%04X, src=%s, dst=%s, id=%d", 
	      i, ptr, ptr->datalen, ptr->flags,
	      srcaddr, dstaddr, ntohs(ptr->header.ip_id));
	while (data) {
	  debug(2, "fragment bucket %d: [%p], data[%p], offset=%u, len=%u", 
		i, ptr, data, data->offset, data->datalen);
	  data = data->next;
	}
      }
      /* cleanup */
      tmp = ptr;
      ptr = ptr->next;
      delete_ip_fragment_data(tmp);
      free(tmp);
    }
  }
}


/* TFTP operations */
#define TFTP_RRQ    1
#define TFTP_WRQ    2
#define TFTP_DATA   3
#define TFTP_ACK    4
#define TFTP_ERROR  5
#define TFTP_OACK   6

/* flags */
#define FILE_OPENED    0x0001
#define FILE_CLOSED  0x0002
#define FILE_FINISHED  0x0004

#define IS_SET(X,Y) ( X & Y )
#define SET_FLAG(X,Y) ( X = Y )

/* linked list */
typedef struct tftp_request {
  struct tftp_request *next;
  u_short op;
  char *filename;
  char *mode;
  u_int32_t src;
  u_int32_t dst;
  u_short sport;
  u_short dport;
  u_short stid;
  u_short dtid;
  u_int blksize;
  u_int timeout;
  size_t tsize;
  u_int last_ack;
  u_int last_block;
  u_char *data;
  u_int datalen;
  size_t len;
  char *name;
  FILE *fp;
  u_int flags;
} tftp_request_t;

typedef struct request_context {
  u_short op;
  u_short errorcode;
  const char *errmsg;
  const char *blksize;
  const char *timeout;
  const char *tsize;
  u_int ack_block;
  u_int block;
  u_int blocksize;
} request_context_t;

enum s_state {
  s_begin = 1,
  s_mode,
  s_filename,
  s_option_blksize,
  s_option_timeout, 
  s_option_tsize,
  s_option_unknown,
  s_option_value,
  s_invalid = 99,
};

int
is_tftp_request(const u_char *packet, u_int caplen, tftp_request_t *request, request_context_t *context)
{

  /* apply heuristic to determine if packet is TFTP request */

  const u_char *ptr;
  const char *str;
  u_short s1, s2;
  u_int c;

  int state = s_begin;

  if (caplen < 9) {
    /* packet too short for TFTP request traffic */
    return 0;
  }

  ptr = packet;
  s1 = *ptr++;
  s2 = *ptr++;

  if ((0 == s1) && (TFTP_RRQ == s2 || TFTP_WRQ == s2)) {
    request->op = s2;
    for(c = 2, str = (const char *)ptr; c < caplen; ++c, ++ptr) {
      if (0 == *ptr) {
	if (2 == c) {
	  /* null at position 2 */
	  return 0;
	}
	switch (state) {
	case s_begin:
	  /* found the filename field */
	  request->filename = (char*)str;
	  state = s_filename;
	  break;
	case s_filename:
	  /* found the mode field */
	  if (
	      (strcasecmp(str, "netascii") == 0) ||
	      (strcasecmp(str, "octet") == 0) ||
	      (strcasecmp(str, "mail") == 0) /* mail is obsolete */
	      ) {
	    request->mode = (char*)str;
	    state = s_mode;
	  } else {
	    state = s_invalid;
	  }
	  break;
	case s_mode:
	case s_option_value:
	  if  (strcasecmp(str, "blksize") == 0) {
	    state = s_option_blksize;
	  } else if (strcasecmp(str, "timeout") == 0) {
	    state = s_option_timeout;
	  } else if (strcasecmp(str, "tsize") == 0) {
	    state = s_option_tsize;
	  } else {
	    state = s_option_unknown;
	  }
	  break;
	case s_option_blksize:
	  context->blksize = str;
	  state = s_option_value;
	  break;
	case s_option_timeout:
	  context->timeout = str;
	  state = s_option_value;
	  break;
	case s_option_tsize:
	  context->tsize = str;
	  state = s_option_value;
	  break;
	case s_option_unknown:
	  /* don't care for now */
	  state = s_option_value;
	  break;
	default:
	  /* invalid */
	  return 0;
	}
	str = (const char *) ptr + 1;
      }
    }
  }
  if (s_mode == state || s_option_value == state) {
    debug(6, "found tftp request; op: %d, filename: %s, mode: %s", request->op, request->filename, request->mode);
    return 1;
  }
  return 0;
}

tftp_request_t * tftp_requests = NULL;

void
cleanup_tftp_request(tftp_request_t *request)
{
  if (NULL != request->fp) {
    fflush(request->fp);
    fclose(request->fp);
    request->fp = NULL;
  }
  if (NULL != request->filename) {
    free(request->filename);
    request->filename = NULL;
  }
  if (NULL != request->mode) {
    free(request->mode);
    request->mode = NULL;
  }
  if (NULL != request->name) {
    free(request->name);
    request->name = NULL;
  }
  if (NULL != request->data) {
    free(request->data);
    request->data = NULL;
  }

}

void
generate_tftp_name(tftp_request_t *request) 
{

  size_t namelen, len;
  u_int32_t s,d;
  char *p1;
  const char *p2;

  if (NULL != request->name) {
    free(request->name);
    request->name = NULL;
  }

  /* base name on src/dst ip&port + filename */
  namelen = 48;
  if (include_filename) {
      len = strlen(request->filename) + 1;
      if (len + namelen < namelen) {
	die(3, "filename too long: %s", request->filename);
      }
      namelen += len;
  }

  if ( (request->name = (char *)malloc(namelen)) == NULL) {
    die(3, "out of memory allocating request->name");
  }

  s = request->src;
  d = request->dst;

  sprintf(request->name, 
	  "%03d.%03d.%03d.%03d.%05d-%03d.%03d.%03d.%03d.%05d",
	  (s >> 24) & 0xff,
	  (s >> 16) & 0xff,
	  (s >>  8) & 0xff,
	  (s >>  0) & 0xff,
	  request->sport,
	  (d >> 24) & 0xff,
	  (d >> 16) & 0xff,
	  (d >>  8) & 0xff,
	  (d >>  0) & 0xff,
	  (request->dtid ? request->dtid : request->dport)
	  );
  if (include_filename) {
    /* append a normalized filename */
    strcat(request->name, "-");
    p1 = request->name + strlen(request->name);
    p2 = request->filename;
    while (*p2) {
      *p1 = ( ('.' == *p2) || isalnum(*p2) ) ? *p2 : '_';
      ++p1;
      ++p2;
    }
    *p1 = '\0';
  }

}

void
close_open_files() 
{
  tftp_request_t *request;
  for (request = tftp_requests; request; request = request->next) {
    if (IS_SET(request->flags, FILE_OPENED)) {
      fflush(request->fp);
      fclose(request->fp);
      request->fp = NULL;
      SET_FLAG(request->flags, FILE_CLOSED);
      break;
    }
  }
}

void
flush_tftp_data(tftp_request_t *request)
{
  int fd, flags;
  char *mode;

  if (NULL == request->data) { 
    return;
  }
  if (IS_SET(request->flags, FILE_FINISHED)) {
    /* all data has been processed */
    return;
  }

  if (console_print) {
    printf("%s: ", request->name);
    fwrite(request->data, sizeof(u_char), request->datalen, stdout);
    putc('\n', stdout);
    fflush(stdout);
  } else {
    if (IS_SET(request->flags, FILE_OPENED)) {
      /* already opened */
    } else {
      flags = 0;
      if (IS_SET(request->flags, FILE_CLOSED)) {
	/* opened, but currently closed */
	flags = O_WRONLY|O_APPEND;
	mode = "ab";
      } else {
	flags = O_WRONLY|O_CREAT;
	if (file_overwrite) {
	  flags |= O_TRUNC;
	} else {
	  flags |= O_EXCL;
	}
	mode = "wb";
      }
      debug(5, "opening file: %s, mode: %s", request->name, mode);
      if ( (fd = open(request->name, flags, 0600)) < 0) {
	if ( (EMFILE == fd) || (ENFILE == fd) ) {
	  /* try to free some file descriptors */
	  close_open_files();
	  fd = open(request->name, flags, 0600);
	}
      }
      if ( (fd < 0) || ( (request->fp = fdopen(fd, mode)) == NULL)) {
	  /* permanent failure */
	  warning("unable to open file [%s]: %s", request->name, strerror(errno));
	  SET_FLAG(request->flags, FILE_FINISHED);
	  cleanup_tftp_request(request);
	  return;
      }
      SET_FLAG(request->flags, FILE_OPENED);
    }
    debug(6, "writing file: %s, datalen: %d, last_block: %d, last_ack: %d", request->name, request->datalen, request->last_block, request->last_ack );
    fwrite(request->data, sizeof(u_char), request->datalen, request->fp);
  }
  
  free(request->data);
  request->data = NULL;

}

void
finalize_outstanding_requests()
{
  /* go through and cleanup any outstanding requests, then remove */
  tftp_request_t *request, *next;

  for (next = tftp_requests; next; ) {
    request = next;
    if (IS_SET(request->flags, FILE_OPENED) || IS_SET(request->flags, FILE_CLOSED)) {
      warning("possible partial content for file: %s",  request->name);
      if (request->last_block == (request->last_ack + 1)) {
	/* didn't receive next ack */
	flush_tftp_data(request);
      }
    }
    cleanup_tftp_request(request);
    next = request->next;
    free(request);
  }
}

void
remove_request_if_exists(tftp_request_t *request)
{
  tftp_request_t *p, *r;
  if (tftp_requests) {
    p = NULL;
    for (r = tftp_requests; r; r = r->next) {
      if (
	  (r->src == request->src) && 
	  (r->dst == request->dst) &&
	  (0xffffffff != request->dst) && /* exclude broadcast */
	  (r->sport == request->sport) &&
	  (r->dport == request->dport)
	  ) {
	/* matched an existing request */
	debug(4, "matched an existing request; %x:%d > %x:%d", r->src, r->sport, r->dst, r->dport);
	if (request->last_block == (request->last_ack + 1)) {
	  /* didn't receive next ack */
	  flush_tftp_data(request);
	}
	cleanup_tftp_request(r);
	if (r == tftp_requests) {
	  tftp_requests = r->next;
	} else {
	  p->next = r->next;
	}
	free(r);
	return;
      } 
      p = r;
    }
  }
  
}

tftp_request_t *
add_tftp_request(tftp_request_t *request)
{

  remove_request_if_exists(request);

  tftp_request_t *newreq = (tftp_request_t *)malloc(sizeof(tftp_request_t));
  if (NULL == newreq) {
    die(2, "failed to malloc tftp_request_t");
  }

  memset(newreq, 0, sizeof(tftp_request_t));

  newreq->op = request->op;
  if (NULL == (newreq->filename = strdup(request->filename))) {
    die(3, "failed to strdup filename");
  }
  if (NULL == (newreq->mode = strdup(request->mode))) {
    die(3, "failed to strdup mode");
  }
  newreq->src = request->src;
  newreq->dst = request->dst;
  newreq->sport = request->sport;
  newreq->dport = request->dport;
  newreq->stid = request->sport;
  newreq->dtid = 0; /* starts unestablished */
  newreq->blksize = 512; /* default */

  debug(4, "created new request [%x]; %x:%d > %x:%d", newreq, 
	newreq->src, newreq->sport, newreq->dst, newreq->dport);

  generate_tftp_name(newreq);
  
  /* always add to front of list, need more specific first to work with broadcast hack */
  newreq->next = tftp_requests;
  tftp_requests = newreq;

  return newreq;
}

void
process_tftp_error(tftp_request_t *request, const u_char *packet, u_int caplen, request_context_t *context)
{
  u_short errorcode, *e;
  const char *errmsg = "";
  const u_char *ptr = packet;

  e = (u_short *) ptr;
  errorcode = ntohs(*e);
  ptr += 2;

  if (caplen > 2) {
    errmsg = (const char *)ptr;
  }
  context->errorcode = errorcode;
  context->errmsg = errmsg;

}

void
process_tftp_ack(tftp_request_t *request, const u_char *packet, u_int caplen, request_context_t *context)
{
  u_short ack_block, *b;
  b = (u_short*) packet;
  ack_block = ntohs(*b);

  context->ack_block = ack_block;

  if (0 == ack_block) {
    if (TFTP_RRQ == request->op) {
      /* extended options */
    }
  } else {
    if (TFTP_RRQ == request->op) {
      if ((request->last_block == ack_block)  && ((request->last_ack + 1) == ack_block)) {
	debug(5, "ack for read; block=%d, last_block=%d, last_ack=%d", 
	      ack_block, request->last_block, request->last_ack);
	flush_tftp_data(request);
	request->last_ack = ack_block;
      } else {
	debug(4, "spurious ack for read; block=%d, last_block=%d, last_ack=%d", 
	      ack_block, request->last_block, request->last_ack);
      }
    } else if (TFTP_WRQ == request->op) {
      if ((request->last_block == ack_block) && ((request->last_ack + 1) == ack_block)) {
	debug(5, "ack for write; block=%d, last_block=%d, last_ack=%d", 
	      ack_block, request->last_block, request->last_ack);
	flush_tftp_data(request);
	request->last_ack = ack_block;
      } else {
	debug(4, "spurious ack for write; block=%d, last_block=%d, last_ack=%d", 
	      ack_block, request->last_block, request->last_ack);
      }
    } else {
      die(3, "corrupted request->op: %d", request->op);
    }
  }
}

void
process_tftp_oack(tftp_request_t *request, const u_char *packet, u_int caplen, request_context_t *context)
{

  const char *str;
  u_int c;
  const u_char *ptr = packet;
  int state = s_begin;

  context->blksize = context->timeout = context->tsize = "";

  str = (const char*) ptr;
  for (c = 0, str = (const char *) ptr; c < caplen; ++c, ++ptr) {
    if (0 == *ptr) {
      switch (state) {
      case s_begin:
      case s_option_value:
	if  (strcasecmp(str, "blksize") == 0) {
	  state = s_option_blksize;
	} else if (strcasecmp(str, "timeout") == 0) {
	  state = s_option_timeout;
	} else if (strcasecmp(str, "tsize") == 0) {
	  state = s_option_tsize;
	} else {
	  state = s_option_unknown;
	}
	break;
      case s_option_blksize:
	context->blksize = str;
	request->blksize = atoi(str);
	if (request->blksize <= 0) {
	  request->blksize = 512;
	}
	state = s_option_value;
	break;
      case s_option_timeout:
	context->timeout = str;
	request->timeout = atoi(str);
	state = s_option_value;
	break;
      case s_option_tsize:
	context->tsize = str;
	request->tsize = atoi(str);
	state = s_option_value;
	break;
      case s_option_unknown:
	state = s_option_value;
	break;
      default:
	/* invalid */
	break;
      }
      str = (const char *) ptr + 1;
    }
  }
  
}

void
process_tftp_data(tftp_request_t *request, const u_char *packet, u_int caplen, request_context_t *context)
{
  u_short block, *b;
  u_int datalen;
  const u_char *ptr = packet;

  b = (u_short*) ptr;
  block = ntohs(*b);
  ptr += 2;
  datalen = caplen - 2;

  context->block = block;
  context->blocksize = datalen;

  if (IS_SET(request->flags, FILE_FINISHED)) {
    return;
  }

  if (NULL != request->data) {
    debug(4, "data already pending on tftp request");
    if (request->datalen < datalen) {
      /* XXX: ??? */
      debug(3, "later request claims more data; request->datalen=%d, datalen=%d", request->datalen, datalen);
      return;
    } else if (0 < datalen) {
      memcpy(request->data, ptr, datalen);
    }
  } else {
    if (0 < datalen) {
      if ( (request->data = (u_char*)malloc(datalen)) == NULL) {
	die(3, "failed to allocate request->data");
      }
      memcpy(request->data, ptr, datalen);
    }
  }

  request->datalen = datalen;
  request->last_block = block;

  if (datalen < request->blksize) {
    /* last block */
    debug(5, "received last block; length: %d, block: %d", datalen, block);
    flush_tftp_data(request);
    SET_FLAG(request->flags, FILE_FINISHED);
    cleanup_tftp_request(request);
  } 

}

int
process_tftp_flow(const u_char *packet, u_int caplen, u_int32_t src, u_int32_t dst, u_short sport, u_short dport, request_context_t *context)
{

  tftp_request_t *request = tftp_requests;
  const u_char *ptr;
  u_short s1, s2, op;

  if (caplen < 4) {
    /* packet too short for TFTP flow traffic */
    return 0;
  }

  /* does the packet look like TFTP ? */
  ptr = packet;
  s1 = *ptr++;
  s2 = *ptr++;

  if (0 != s1) {
    return 0;
  }

  switch (s2) {
  case TFTP_ACK:
  case TFTP_OACK:
  case TFTP_DATA:
  case TFTP_ERROR:
    op = s2;
    break;
  default:
    return 0;
  }

  context->op = op;

  /* search existing requests */
 again:
  for (request = tftp_requests; request; request = request->next) {
    
    /* is this broadcast -> client ? */
    if (
	(0xffffffff == request->dst) &&
	(dst == request->src) &&
	(dport == request->stid)
	) {
      /* create a fake request for this */
      tftp_request_t *fakereq = add_tftp_request(request);
      fakereq->dst = src;
      goto again;
    } else if ( /* is this server -> client ? */
	(src == request->dst) && 
	(dst == request->src) &&
	(dport == request->stid)
	) {
      if (0 != request->dtid) {
	/* already seen */
	if (request->dtid != sport) {
	  debug(2, "found matching address, but dtid wrong; src=%x, dst=%x, dport=%d, sport=%d, dtid=%d", 
		src, dst, dport, sport, request->dtid);
	  continue;
	}
      } else {
	/* first response packet from server store TID */
	request->dtid = sport;
	/* regenerate tftp name */
	generate_tftp_name(request);
      }

      if (TFTP_ERROR == op) {
	process_tftp_error(request, ptr, caplen - 2, context);
      } else if ((TFTP_ACK == op) && (TFTP_WRQ == request->op)) {
	process_tftp_ack(request, ptr, caplen - 2, context);
      } else if ((TFTP_OACK == op) && (TFTP_WRQ == request->op)) {
	process_tftp_oack(request, ptr, caplen - 2, context);
      } else if ((TFTP_OACK == op) && (TFTP_RRQ == request->op)) {
	process_tftp_oack(request, ptr, caplen - 2, context);
      } else if ((TFTP_DATA == op) && (TFTP_RRQ == request->op)) {
	process_tftp_data(request, ptr, caplen - 2, context);
      } else {
	/* invalid state */
	  debug(2, "reached invalid/unhandled server->client state; src=%x, dst=%x, dport=%d, sport=%d, dtid=%d, op=%d, request->op=%d", 
		src, dst, dport, sport, request->dtid, op, request->op);
	  return 0;
      }

      return 1;

    } else if ( /* is this client -> server */
	       (src == request->src) && 
	       (dst == request->dst) &&
	       (dport == request->dtid) && 
	       (sport == request->stid)
	       ) {
      if (TFTP_ERROR == op) {
	process_tftp_error(request, ptr, caplen - 2, context);
      } else if ((TFTP_ACK == op) && (TFTP_RRQ == request->op)) {
	process_tftp_ack(request, ptr, caplen - 2, context);
      } else if ((TFTP_DATA == op) && (TFTP_WRQ == request->op)) {
	process_tftp_data(request, ptr, caplen - 2, context);
      } else {
	/* invalid state */
	debug(2, "reached invalid/unhandled client->server state; src=%x, dst=%x, dport=%d, sport=%d, dtid=%d, op=%d, request->op=%d", 
	      src, dst, dport, sport, request->dtid, op, request->op);
	return 0;
      }

      return 1;
    } else if (/* is this unestablished client -> server */
	       (src == request->src) && 
	       (dst == request->dst) &&
	       (0 == request->dtid) && 
	       (sport == request->stid)
	       ) {
      if (TFTP_ERROR == op) {
	process_tftp_error(request, ptr, caplen - 2, context);
	return 1;
      } else {
	/* invalid state */
	debug(2, "reached invalid/unhandled unestablished client->server state; src=%x, dst=%x, dport=%d, sport=%d, dtid=%d, op=%d, request->op=%d", 
	      src, dst, dport, sport, request->dtid, op, request->op);
	return 0;
      }
    }
  }

  return 0;
}

void
process_udp(const u_char *packet, u_int caplen, const struct ip *iph)
{
  u_int udphsize = sizeof(struct udphdr);
  const struct udphdr *udph = (struct udphdr *) packet;
  char srcaddr[INET_ADDRSTRLEN], dstaddr[INET_ADDRSTRLEN];
  u_int ulen, sport, dport;
  const u_char *payload;
  u_int payload_len = caplen - 8;
  tftp_request_t request;
  request_context_t context;
  u_int32_t src, dst;
  u_short cksum = 0;

  if (caplen < udphsize) {
    warning("udp header is not valid; %d bytes, but should be at least %d; id=%d", 
	    caplen, udphsize, ntohs(iph->ip_id));
    return;
  }

  src = iph->ip_src.s_addr;
  dst = iph->ip_dst.s_addr;

  ulen = ntohs(udph->uh_ulen);
  sport = ntohs(udph->uh_sport);
  dport = ntohs(udph->uh_dport);

  if (payload_len != (ulen - 8)) { 
    warning("invalid udp header; specified length: %d, doesn't match expected: %d; id=%d",
	    ulen - 8, payload_len, ntohs(iph->ip_id));
    return;
  }
  payload = packet + 8;

  memset(&request, 0, sizeof(tftp_request_t));
  memset(&context, 0, sizeof(request_context_t));

  /* valid checksum ? */
  if (0 != udph->uh_sum) {
    cksum = udp_checksum(iph, udph, caplen);
    if (0 != cksum) {
      if (check_bad_checksums) {
	if (inet_ntop(AF_INET, &src, srcaddr, sizeof(srcaddr)) == NULL) {
	  strcpy(srcaddr, "?");
	}
	if (inet_ntop(AF_INET, &dst, dstaddr, sizeof(dstaddr)) == NULL) {
	  strcpy(dstaddr, "?");
	}
	warning("invalid udp checksum; checksum=%x; src=%s:%d, dst=%s:%d, id=%d", 
		cksum, srcaddr, sport, dstaddr, dport, ntohs(iph->ip_id));
	return;
      } else {
	debug(3, "allowing invalid udp checksum; checksum=%x", cksum);
      }
    }
  }

  if (is_tftp_request(payload, payload_len, &request, &context)) {
    request.src = ntohl(src);
    request.dst = ntohl(dst);
    request.sport = sport;
    request.dport = dport;
    add_tftp_request(&request);
    if (verbose) {
      if (inet_ntop(AF_INET, &src, srcaddr, sizeof(srcaddr)) == NULL) {
	strcpy(srcaddr, "?");
      }
      if (inet_ntop(AF_INET, &dst, dstaddr, sizeof(dstaddr)) == NULL) {
	strcpy(dstaddr, "?");
      }
      printf("%-5s %s:%d > %s:%d \"%s\" %s", 
	     (TFTP_RRQ == request.op ? "READ" : "WRITE"),
	     srcaddr, sport, dstaddr, dport, request.filename, request.mode);
      if (context.blksize) { printf(" blksize=%s", context.blksize); }
      if (context.timeout) { printf(" timeout=%s", context.timeout); }
      if (context.tsize) { printf(" tsize=%s", context.tsize); }
      putc('\n', stdout);
      if (dump_packet) {
	hexdump(payload, payload_len);
      }
      fflush(stdout);
    }
  } else if (process_tftp_flow(payload, payload_len, ntohl(src), ntohl(dst), sport, dport, &context)) {
    if (verbose > 1) {
      if (inet_ntop(AF_INET, &src, srcaddr, sizeof(srcaddr)) == NULL) {
	strcpy(srcaddr, "?");
      }
      if (inet_ntop(AF_INET, &dst, dstaddr, sizeof(dstaddr)) == NULL) {
	strcpy(dstaddr, "?");
      }
      switch(context.op) {
      case TFTP_ACK:
	printf("%-5s %s:%d > %s:%d %d\n", 
	       "ACK",
	       srcaddr, sport, dstaddr, dport, context.ack_block);
	break;
      case TFTP_OACK:
	printf("%-5s %s:%d > %s:%d", 
	       "OACK",
	       srcaddr, sport, dstaddr, dport);
	if (context.blksize) { printf(" blksize=%s", context.blksize); }
	if (context.timeout) { printf(" timeout=%s", context.timeout); }
	if (context.tsize) { printf(" tsize=%s", context.tsize); }
	putc('\n', stdout);
	break;
      case TFTP_DATA:
	printf("%-5s %s:%d > %s:%d %d %d\n", 
	       "DATA",
	       srcaddr, sport, dstaddr, dport, context.block, context.blocksize);
	break;
      case TFTP_ERROR:
	printf("%-5s %s:%d > %s:%d %d \"%s\"\n", 
	       "ERROR",
	       srcaddr, sport, dstaddr, dport, context.errorcode, context.errmsg);
	break;
      }
      if (dump_packet) {
	hexdump(payload, payload_len);
      }
      fflush(stdout);
    }
  } else if (verbose > 2) {
    /* does not appear related to TFTP */
    if (inet_ntop(AF_INET, &src, srcaddr, sizeof(srcaddr)) == NULL) {
      strcpy(srcaddr, "?");
    }
    if (inet_ntop(AF_INET, &dst, dstaddr, sizeof(dstaddr)) == NULL) {
      strcpy(dstaddr, "?");
    }
    printf("%-5s %s:%d > %s:%d \n", 
	   "XXX",
	   srcaddr, sport, dstaddr, dport);
    if (dump_packet) {
      hexdump(payload, payload_len);
    }
    fflush(stdout);
  }
}

void
process_ip(const u_char *packet, u_int caplen)
{
  u_int iphsize = sizeof(struct ip);
  const struct ip *iph = (struct ip *) packet;
  u_int len;
  u_int header_len;
  u_int offset;
  u_short morefrags = 0;
  u_short cksum = 0;

  if (caplen < iphsize) {
    warning("ip header is not valid; %d bytes, but should be at least", caplen, iphsize);
    return;
  }

  if (4 != iph->ip_v) {
    warning("unexpected version %d; not 4",iph->ip_v);
    return;
  }

  if (IPPROTO_UDP != iph->ip_p) {
    warning("unexpected protocol %d; not udp", iph->ip_p);
    return;
  }

  len = ntohs(iph->ip_len);
  if (caplen < len) {
    warning("partial ip packet; length: %d, captured: %d; id=%d", len, caplen, ntohs(iph->ip_id));
    return;
  }

  header_len = iph->ip_hl << 2;
  if (len < header_len) {
    warning("partial ip packet; length: %d, header length: %d; id=%d", 
	    len, header_len, ntohs(iph->ip_id));
    return;
  }

  if ( (cksum = checksum(packet, header_len, 0)) != 0) {
    if (check_bad_checksums) {
      warning("invalid ip checksum; checksum=%x; id=%d", cksum, ntohs(iph->ip_id));
      return;
    } else {
      debug(3, "allowing invalid ip checksum; checksum=%x", cksum);
    }
  }

  /* check for fragmented ip */
  morefrags = ntohs(iph->ip_off) & IP_MF;
  offset = ntohs(iph->ip_off) & IP_OFFMASK;
  if (morefrags || offset) {
    ip_fragment_t *frag;
    frag = add_ip_fragment(iph, offset, morefrags, packet + header_len, len - header_len);
    if ((NULL != frag) && (ip_fragment_complete(frag))) {
      process_udp(frag->packetdata, frag->datalen, &(frag->header));
      remove_ip_fragment(frag);
    }
  } else {
    process_udp(packet + header_len, len - header_len, iph);
  }
}

void 
process_ethernet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
  u_int len = header->len;
  u_int caplen = header->caplen;
  u_int ethsize = sizeof(struct ether_header);
  u_short ether_type;
  u_int ieee802size = ethsize + 8;
  const struct ether_header *eh = (struct ether_header*) packet;

  if (caplen < ethsize) {
    warning("ethernet is not valid; %d bytes, but should be at least", caplen, ethsize);
    return;
  }

  if (caplen != len) {
    if (1518 == len && caplen >= 1514) {
      /* workaround for short captures on max ethernet frames */
      caplen = 1514;
    } else if (caplen < len) {
      warning("only %d bytes of %d bytes were captured", caplen, len);
      return;
    } else { 
      caplen = len;
    }
  }

  ether_type = htons(eh->ether_type);

  if (ETHERTYPE_IP == ether_type) {
    process_ip(packet + ethsize, caplen - ethsize);
    return;
  }

  /* check for ieee802 */
  if (caplen >= ieee802size) {
    ether_type = htons(*(short *)(packet + ieee802size - 2));
    if (ETHERTYPE_IP == ether_type) {
      process_ip(packet + ieee802size, caplen - ieee802size);
      return;
    }
  }

  warning("ethernet has wrong type: %d", htons(eh->ether_type));
  
}

void 
process_raw(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
  u_int len = header->len;
  u_int caplen = header->caplen;

  if (caplen != len) {
    warning("only %d bytes of %d bytes were captured", caplen, len);
    return;
  }

  process_ip(packet, caplen);

}

void 
process_linux_sll(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
  u_int len = header->len;
  u_int caplen = header->caplen;
  u_int linux_sllsize = 16;
  u_short proto;

  if (caplen != len) {
    warning("only %d bytes of %d bytes were captured", caplen, len);
    return;
  }

  if (caplen < linux_sllsize) {
    warning("linux_sll is not valid; %d bytes, but should be at least", 
	    caplen, linux_sllsize);
    return;
  }

  process_ip(packet + linux_sllsize, caplen - linux_sllsize);
  
}

void 
process_ppp(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
  u_int len = header->len;
  u_int caplen = header->caplen;
  u_int pppsize = 4;
  u_short proto;

  if (caplen != len) {
    warning("only %d bytes of %d bytes were captured", caplen, len);
    return;
  }

  if (caplen < pppsize) {
    warning("ppp is not valid; %d bytes, but should be at least", caplen, pppsize);
    return;
  }

  proto = htons(*(u_short*)(packet + 2));

  if (0x0021 == proto) {
    process_ip(packet + pppsize, caplen - pppsize);
  }
  
}

PACKET_HANDLER
find_handler(int dltype)
{
  int i = 0;
  const char *dldesc;

  struct {
    PACKET_HANDLER handler;
    int dltype;
  } handlers[] = 
    { 
      {process_ethernet, DLT_EN10MB}, 
      {process_ethernet, DLT_IEEE802},
      {process_raw, DLT_NULL},
      {process_ppp, DLT_PPP},
#ifdef DLT_RAW
      {process_raw, DLT_RAW},
#endif
#ifdef DLT_LINUX_SLL
      {process_linux_sll, DLT_LINUX_SLL},
#endif
      {0 , 0} 
    };

  while (handlers[i].handler) {
    if (handlers[i].dltype == dltype) {
      return handlers[i].handler;
    }
    ++i;
  }

#if HAVE_PCAP_DATALINK_VAL_TO_DESCRIPTION
  dldesc = pcap_datalink_val_to_description(dltype);
#else
  dldesc = NULL;
#endif

  if (NULL == dldesc) {
    die(2, "unable to find handler for type %d", dltype);
  } else {
    die(2, "unable to find handler for type %d (%s)", dltype, dldesc);
  }

  return NULL;
}

int
main(int argc, char **argv)
{

  int rc = -1, do_usage = 0;
  int ch, dl, i;
  struct stat sb;
  char *pcapfile = "-";
  pcap_t *handle = NULL;
  char errbuf[PCAP_ERRBUF_SIZE];
  struct bpf_program fp;
  char default_filter[] = "udp";
  char *filter_exp = default_filter;
  size_t exprlen, len;
  int dltype;
  const char *dlname, *dldesc;

  PACKET_HANDLER packet_handler;

  memset(ip_fragments, 0, sizeof(ip_fragments));

  while ( (ch = getopt(argc, argv, "vr:hd:XcfEB")) != -1 ) {
    switch (ch) {
    case 'r':
      pcapfile = optarg;
      if (0 != strcmp(pcapfile, "-")) { 
	if ( (stat(pcapfile, &sb)) < 0) {
	  die(2, "unable to access file %s: %s", pcapfile, strerror(errno));
	}
      }
      break;
    case 'd':
      dl = atoi(optarg);
      if (dl < 0) {
	warning("bad debug-level: %d; setting to %d", dl, debuglevel);
      } else {
	debuglevel = dl;
      }
      break;
    case 'v':
      verbose += 1;
      break;
    case 'h':
      do_usage = 1;
      rc = 0;
      break;
    case 'X':
      dump_packet = 1;
      break;
    case 'c':
      console_print = 1;
      break;
    case 'f':
      file_overwrite = 1;
      break;
    case 'E':
      include_filename = 0;
      break;
    case 'B':
      check_bad_checksums = 1;
      break;
    case '?':
    default:
      rc = do_usage = 1;
    }
  }

  if (do_usage) {
    usage(argv[0]);
    exit(rc);
  }
  rc = 0;

  argc -= optind;
  argv += optind;

  if (argc > 0) {
    exprlen = strlen(filter_exp) + 5;
    for (i = 0; i < argc; ++i) {
      len = strlen(argv[i]) + 1;
      if (len + exprlen < exprlen) {
	die(3, "expression too long");
      }
      exprlen += len;
    }
    exprlen += 1;
    filter_exp = (char *)malloc(exprlen);
    if (NULL == filter_exp) {
      die(2, "unable to allocate memory for filter expression");
    }
    strcpy(filter_exp, default_filter);
    strcat(filter_exp, " and ");
    for (i = 0; i < argc; ++i) {
      strcat(filter_exp, argv[i]);
      if ((i + 1) < argc) {
	strcat(filter_exp, " ");
      }
    }
  }

  if ( (handle = pcap_open_offline(pcapfile, errbuf)) == NULL) {
    die(2, "unable to open pcap file %s: %s", pcapfile, errbuf);
  }
  
  dltype = pcap_datalink(handle);
  packet_handler = find_handler(dltype);

#if HAVE_PCAP_DATALINK_VAL_TO_DESCRIPTION
  dldesc = pcap_datalink_val_to_description(dltype);
#else
  dldesc = NULL;
#endif
#if HAVE_PCAP_DATALINK_VAL_TO_NAME
  dlname = pcap_datalink_val_to_name(dltype);
#else
  dlname = NULL;
#endif

  if ( (NULL == dlname) || (NULL == dldesc) ) {
    fprintf(stderr, "reading from file %s, using datalink type %d\n",  
	    pcapfile, dltype);
  } else {
    fprintf(stderr, "reading from file %s, using datalink type %s (%s)\n",  
	    pcapfile, dlname, dldesc);
  }

  if (pcap_compile(handle, &fp, filter_exp, 0, 0) < 0) {
    die(2, "unable to compile pcap filter %s: %s", filter_exp, pcap_geterr(handle));
  }

  if (pcap_setfilter(handle, &fp) < 0) {
    die(2, "unable to set pcap filter %s: %s", filter_exp, pcap_geterr(handle));
  }

  if (pcap_loop(handle, -1, packet_handler, 0) < 0) {
    rc = 1;
    die(0, "failed processing packets: %s", pcap_geterr(handle));
  }

  pcap_close(handle);

  finalize_outstanding_requests();

  cleanup_ip_fragments();

  if (filter_exp != default_filter) {
    free(filter_exp);
  }
     
  exit(rc);
}
